from abc import ABC, abstractmethod
from inspect import isabstract
from typing import Set, TypeVar

from humps import kebabize

from kilroy_server_py_utils.utils import classproperty

T = TypeVar("T")


def _get_all_subclasses(cls: T) -> Set[T]:
    all_subclasses = set()

    for subclass in cls.__subclasses__():
        all_subclasses.add(subclass)
        all_subclasses.update(_get_all_subclasses(subclass))

    return all_subclasses


class Categorizable(ABC):
    @classproperty
    @abstractmethod
    def category(cls) -> str:
        pass

    @classproperty
    def pretty_category(cls) -> str:
        return kebabize(cls.category).replace("-", " ").title()

    @classmethod
    def for_category(cls: T, category: str) -> T:
        for categorizable in cls.all_categorizables:
            if categorizable.category == category:
                return categorizable
        raise ValueError(f'Categorizable for category "{category}" not found.')

    @classproperty
    def all_categorizables(cls: T) -> Set[T]:
        subclasses = _get_all_subclasses(cls)
        return {
            subclass for subclass in subclasses if not isabstract(subclass)
        }

    def __eq__(self, other):
        return (
            isinstance(other, Categorizable)
            and self.category == other.category
        )

    def __hash__(self) -> int:
        return hash(self.category)



import asyncio


class Read(asyncio.Future):
    @staticmethod
    def is_compatible(holds):
        return not holds[Write]


class Write(asyncio.Future):
    @staticmethod
    def is_compatible(holds):
        return not holds[Read] and not holds[Write]



from kilroy_server_py_utils.categorizable import Categorizable
from kilroy_server_py_utils.configurable import Configurable, Configuration
from kilroy_server_py_utils.loadable import (
    EventBasedObservableWrapper,
    Loadable,
    ValueWrapper,
)
from kilroy_server_py_utils.locks import Read, Write
from kilroy_server_py_utils.observable import (
    FetchOnlyObservableWrapper,
    FetchableObservable,
    NotInitializedError,
    Observable,
    ReadOnlyObservableWrapper,
    ReadableObservable,
)
from kilroy_server_py_utils.parameters.base import (
    OptionalParameter,
    Parameter,
    ParameterGetError,
    ParameterSetError,
)
from kilroy_server_py_utils.parameters.categorizable import (
    CategorizableBasedOptionalParameter,
    CategorizableBasedParameter,
)
from kilroy_server_py_utils.parameters.nested import (
    NestedOptionalParameter,
    NestedParameter,
)
from kilroy_server_py_utils.resources import (
    resource,
    resource_bytes,
    resource_text,
)
from kilroy_server_py_utils.savable import Savable
from kilroy_server_py_utils.schema import JSONSchema
from kilroy_server_py_utils.utils import (
    SelfDeletingDirectory,
    background,
    base64_decode,
    base64_encode,
    classproperty,
    noop,
    normalize,
)



import inspect
import json
from contextlib import asynccontextmanager
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
)

from kilroy_server_py_utils.categorizable import Categorizable
from kilroy_server_py_utils.loadable import Loadable
from kilroy_server_py_utils.observable import (
    Observable,
    ReadOnlyObservableWrapper,
    ReadableObservable,
)
from kilroy_server_py_utils.parameters.base import Parameter
from kilroy_server_py_utils.savable import Savable
from kilroy_server_py_utils.schema import JSONSchema
from kilroy_server_py_utils.utils import classproperty, get_generic_args

T = TypeVar("T")
StateType = TypeVar("StateType")
ConfigurationType = TypeVar("ConfigurationType", bound="Configuration")
ConfigurableType = TypeVar("ConfigurableType", bound="Configurable")
CategorizableType = TypeVar("CategorizableType", bound=Categorizable)
SavableType = TypeVar("SavableType", bound=Savable)


class Configuration(Generic[StateType]):
    _parameters: Set[Parameter]
    _state: Loadable[StateType]
    _json: Observable[Dict[str, Any]]

    def __init__(
        self,
        parameters: Set[Parameter],
        state: Loadable[StateType],
        json_: Observable[Dict[str, Any]],
    ) -> None:
        self._parameters = parameters
        self._state = state
        self._json = json_

    @classmethod
    async def build(
        cls: Type[ConfigurationType], parameters: Set[Parameter], **kwargs
    ) -> ConfigurationType:
        return cls(
            parameters=parameters,
            state=await Loadable.build(),
            json_=await Observable.build(),
            **kwargs,
        )

    @property
    def state(self) -> Loadable[StateType]:
        return self._state

    @property
    def json(self) -> ReadableObservable[Dict[str, Any]]:
        return ReadOnlyObservableWrapper(self._json)

    async def _build_json(self) -> Dict[str, Any]:
        state = await self._state.value.fetch()
        return {
            parameter.name: await parameter.get(state)
            for parameter in self._parameters
        }

    @asynccontextmanager
    async def load(
        self,
    ) -> AsyncIterator[Callable[[StateType], Awaitable[None]]]:
        async with self._state.load() as (_, setter):
            yield setter
        await self._json.set(await self._build_json())

    async def set(self, config: Dict[str, Any]) -> Dict[str, Any]:
        params_map = {
            parameter.name: parameter for parameter in self._parameters
        }

        async with self._state.load() as (getter, setter):
            state = await getter()
            undos = []

            try:
                for name, value in config.items():
                    parameter = params_map.get(name, None)
                    if parameter is not None:
                        undo = await parameter.set(state, value)
                        undos.append(undo)
                await setter(state)
            except Exception as e:
                for undo in reversed(undos):
                    try:
                        await undo()
                    except Exception:
                        pass
                raise e

        config = await self._build_json()
        await self._json.set(config)
        return config


class Configurable(Savable, Generic[StateType]):
    _config: Configuration[StateType]

    def __init__(self, config: Configuration[StateType], **kwargs) -> None:
        self._config = config
        self._kwargs = kwargs

    @classmethod
    async def build(cls: Type[ConfigurableType], **kwargs) -> ConfigurableType:
        return cls(config=await Configuration.build(cls.parameters), **kwargs)

    @classmethod
    async def create(
        cls: Type[ConfigurableType], **kwargs
    ) -> ConfigurableType:
        instance = await cls.build(**kwargs)
        await instance.init()
        return instance

    @property
    def config(self) -> Configuration[StateType]:
        return self._config

    @property
    def state(self) -> Loadable[StateType]:
        return self._config.state

    @classmethod
    async def _build_configurable(
        cls, configurable: Type[ConfigurableType], **kwargs
    ) -> ConfigurableType:
        instance = await configurable.build(**kwargs)
        await instance.init()
        return instance

    @classmethod
    async def _build_categorizable(
        cls, categorizable: Type[CategorizableType], category: str, **kwargs
    ) -> CategorizableType:
        subclass = categorizable.for_category(category)
        if issubclass(subclass, Configurable):
            # noinspection PyTypeChecker
            return await cls._build_configurable(subclass, **kwargs)
        return subclass(**kwargs)

    @classmethod
    async def _build_generic(cls, type_: Type[T], **kwargs) -> T:
        if issubclass(type_, Categorizable):
            return await cls._build_categorizable(type_, **kwargs)
        if issubclass(type_, Configurable):
            return await cls._build_configurable(type_, **kwargs)
        return type_(**kwargs)

    @classproperty
    def _state_class(cls) -> Type[StateType]:
        return get_generic_args(cls, Configurable)[0]

    async def _build_default_state(self) -> StateType:
        return self._state_class(**self._kwargs)

    async def init(self) -> None:
        async with self._config.load() as setter:
            state = await self._build_default_state()
            await setter(state)

    async def cleanup(self) -> None:
        pass

    @classmethod
    async def _save_state(cls, state: StateType, directory: Path) -> None:
        with open(directory / "state.json", "w") as f:
            json.dump(vars(state), f)

    async def save(self, directory: Path) -> None:
        directory.mkdir(parents=True, exist_ok=True)
        async with self.state.read_lock() as state:
            await self._save_state(state, directory)

    @classmethod
    async def from_saved(
        cls: Type[ConfigurableType], directory: Path, **kwargs
    ) -> ConfigurableType:
        instance = await cls.build(**kwargs)
        await instance.load_saved(directory)
        return instance

    @classmethod
    async def _load_savable(
        cls, directory: Path, savable: Type[SavableType], **kwargs
    ) -> SavableType:
        return await savable.from_saved(directory, **kwargs)

    @classmethod
    async def _load_categorizable(
        cls,
        directory: Path,
        categorizable: Type[CategorizableType],
        category: str,
        **kwargs,
    ) -> CategorizableType:
        subclass = categorizable.for_category(category)
        if issubclass(subclass, Savable):
            # noinspection PyTypeChecker
            return await subclass.from_saved(directory, **kwargs)
        return await cls._build_categorizable(
            categorizable, category, **kwargs
        )

    @classmethod
    async def _load_generic(
        cls,
        directory: Path,
        type_: Type[T],
        default: Optional[Callable[[], Awaitable[T]]] = None,
        **kwargs,
    ) -> T:
        try:
            if issubclass(type_, Categorizable):
                return await cls._load_categorizable(
                    directory, type_, **kwargs
                )
            if issubclass(type_, Savable):
                return await cls._load_savable(directory, type_, **kwargs)
        except Exception:
            pass

        if default is not None:
            return await default()

        return await cls._build_generic(type_, **kwargs)

    async def _load_saved_state(self, directory: Path) -> StateType:
        with open(directory / "state.json", "r") as f:
            state_dict = json.load(f)
        return self._state_class(**state_dict)

    async def load_saved(self, directory: Path) -> None:
        async with self._config.load() as setter:
            state = await self._load_saved_state(directory)
            await setter(state)

    @classproperty
    def parameters(cls) -> Set[Parameter]:
        return {
            inner_cls()
            for inner_cls in cls.__dict__.values()
            if inspect.isclass(inner_cls) and issubclass(inner_cls, Parameter)
        }

    @classproperty
    def schema_name(cls) -> str:
        return f"{cls.__name__} configuration schema"

    @classproperty
    def schema(cls) -> JSONSchema:
        # noinspection PyArgumentList
        return JSONSchema(
            title=cls.schema_name,
            type="object",
            required=cls.required_properties,
            properties=cls.properties_schema,
        )

    @classproperty
    def required_properties(cls) -> List[str]:
        return [
            parameter.name
            for parameter in cls.parameters
            if parameter.required
        ]

    @classproperty
    def properties_schema(cls) -> Dict[str, Any]:
        return {
            parameter.name: parameter.schema for parameter in cls.parameters
        }



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Type, TypeVar

SavableType = TypeVar("SavableType", bound="Savable")


class Savable(ABC):
    @abstractmethod
    async def save(self, directory: Path) -> None:
        pass

    @classmethod
    @abstractmethod
    async def from_saved(
        cls: Type[SavableType], directory: Path, **kwargs
    ) -> SavableType:
        pass



from abc import ABC
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Optional,
    Type,
    TypeVar,
)

from humps import decamelize

from kilroy_server_py_utils.configurable import Configurable
from kilroy_server_py_utils.parameters.base import OptionalParameter, Parameter
from kilroy_server_py_utils.utils import (
    SelfDeletingDirectory,
    classproperty,
    get_generic_args,
    noop,
)

StateType = TypeVar("StateType")
ParameterType = TypeVar("ParameterType")
ConfigurableType = TypeVar("ConfigurableType", bound=Configurable)


class NestedParameter(
    Parameter[StateType, Dict[str, Any]],
    ABC,
    Generic[StateType, ConfigurableType],
):
    async def _get(self, state: StateType) -> Dict[str, Any]:
        configurable = await self._get_configurable(state)
        return await configurable.config.json.fetch()

    async def _set(
        self, state: StateType, value: Dict[str, Any]
    ) -> Callable[[], Awaitable]:
        configurable = await self._get_configurable(state)
        undo = await self._create_undo(configurable)
        await configurable.config.set(value)
        return undo

    @staticmethod
    async def _create_undo(
        configurable: Configurable,
    ) -> Callable[[], Awaitable]:
        original_config = await configurable.config.json.get()

        async def undo():
            await configurable.config.set(original_config)

        return undo

    async def _get_configurable(self, state: StateType) -> ConfigurableType:
        return getattr(state, decamelize(self.name))

    @classproperty
    def configurable_class(cls) -> Type[ConfigurableType]:
        return get_generic_args(cls, NestedParameter)[1]

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": "object",
            "title": cls.pretty_name,
            "required": cls.configurable_class.required_properties,
            "properties": cls.configurable_class.properties_schema,
        }


class NestedOptionalParameter(
    OptionalParameter[StateType, Dict[str, Any]],
    ABC,
    Generic[StateType, ConfigurableType],
):
    async def _get(self, state: StateType) -> Optional[Dict[str, Any]]:
        configurable = await self._get_configurable(state)
        if configurable is None:
            return None
        return await configurable.config.json.fetch()

    async def _set(
        self, state: StateType, value: Optional[Dict[str, Any]]
    ) -> Callable[[], Awaitable]:
        configurable = await self._get_configurable(state)

        if value is None:
            undo = await self._create_undo(state, configurable, None)
            await self._set_configurable(state, None)
            if configurable is not None:
                await configurable.cleanup()
            return undo

        if configurable is None:
            params = await self._get_params(state)
            new = await self.configurable_class.create(**params)
            await new.config.set(value)
            undo = await self._create_undo(state, None, new)
            await self._set_configurable(state, new)
            return undo

        undo = await self._create_undo(state, configurable, configurable)
        await configurable.config.set(value)
        return undo

    async def _create_undo(
        self,
        state: StateType,
        old: Optional[Configurable],
        new: Optional[Configurable],
    ) -> Callable[[], Awaitable]:
        if old is None and new is None:
            return noop

        if old is new:
            config = await old.config.json.get()

            async def undo():
                await old.config.set(config)

            return undo

        if old is None:

            async def undo():
                await self._set_configurable(state, None)
                await new.cleanup()

            return undo

        tempdir = SelfDeletingDirectory()
        await old.save(tempdir.path)
        params = await self._get_params(state)

        async def undo():
            loaded = await self.configurable_class.from_saved(
                tempdir.path, **params
            )
            await self._set_configurable(state, loaded)
            if new is not None:
                await new.cleanup()

        return undo

    async def _get_configurable(
        self, state: StateType
    ) -> Optional[ConfigurableType]:
        return getattr(state, decamelize(self.name))

    async def _set_configurable(
        self, state: StateType, value: Optional[ConfigurableType]
    ) -> None:
        setattr(state, decamelize(self.name), value)

    async def _get_params(self, state: StateType) -> Dict[str, Any]:
        return getattr(state, f"{decamelize(self.name)}_params")

    @classproperty
    def configurable_class(cls) -> Type[ConfigurableType]:
        return get_generic_args(cls, NestedOptionalParameter)[1]

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": ["object", "null"],
            "title": cls.pretty_name,
            "required": cls.configurable_class.required_properties,
            "properties": cls.configurable_class.properties_schema,
        }



from abc import ABC
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Optional,
    Type,
    TypeVar,
)

from humps import decamelize

from kilroy_server_py_utils.categorizable import Categorizable
from kilroy_server_py_utils.configurable import Configurable
from kilroy_server_py_utils.parameters.base import Parameter
from kilroy_server_py_utils.utils import (
    SelfDeletingDirectory,
    classproperty,
    get_generic_args,
    noop,
)

StateType = TypeVar("StateType")
CategorizableType = TypeVar("CategorizableType", bound=Categorizable)


class CategorizableBasedParameter(
    Parameter[StateType, Dict[str, Any]],
    ABC,
    Generic[StateType, CategorizableType],
):
    async def _get(self, state: StateType) -> Dict[str, Any]:
        categorizable = await self._get_categorizable(state)
        category = categorizable.category
        if isinstance(categorizable, Configurable):
            return {
                "type": category,
                "config": await categorizable.config.json.get(),
            }
        return {"type": category}

    async def _set(
        self, state: StateType, value: Dict[str, Any]
    ) -> Callable[[], Awaitable]:
        current = await self._get_categorizable(state)
        new_category = value.get("type", current.category)

        if new_category == current.category:
            undo = await self._create_undo(state, current, current)
            await current.config.set(value.get("config", {}))
            return undo

        params = await self._get_params(state, new_category)
        subclass = self.categorizable_base_class.for_category(new_category)
        if issubclass(subclass, Configurable):
            instance = await subclass.create(**params)
            await instance.config.set(value.get("config", {}))
        else:
            instance = subclass(**params)

        undo = await self._create_undo(state, current, instance)
        await self._set_categorizable(state, instance)
        if isinstance(current, Configurable):
            await current.cleanup()
        return undo

    async def _create_undo(
        self, state: StateType, old: Categorizable, new: Categorizable
    ) -> Callable[[], Awaitable]:
        if old is new:
            if not isinstance(old, Configurable):
                return noop

            config = await old.config.json.get()

            async def undo():
                # noinspection PyUnresolvedReferences
                await old.config.set(config)

            return undo

        if isinstance(old, Configurable):
            tempdir = SelfDeletingDirectory()
            await old.save(tempdir.path)
            # noinspection PyUnresolvedReferences
            params = await self._get_params(state, old.category)

            async def undo():
                loaded = await old.from_saved(tempdir.path, **params)
                await self._set_categorizable(state, loaded)
                if isinstance(new, Configurable):
                    await new.cleanup()

            return undo

        async def undo():
            await self._set_categorizable(state, old)
            if isinstance(new, Configurable):
                await new.cleanup()

        return undo

    async def _get_categorizable(self, state: StateType) -> CategorizableType:
        return getattr(state, decamelize(self.name))

    async def _set_categorizable(
        self, state: StateType, value: CategorizableType
    ) -> None:
        setattr(state, decamelize(self.name), value)

    async def _get_params(
        self, state: StateType, category: str
    ) -> Dict[str, Any]:
        all_params = getattr(state, f"{decamelize(self.name)}s_params")
        return all_params.get(category, {})

    @classproperty
    def categorizable_base_class(cls) -> Type[CategorizableType]:
        return get_generic_args(cls, CategorizableBasedParameter)[1]

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        options = []
        for categorizable in cls.categorizable_base_class.all_categorizables:
            properties = {
                "type": {
                    "type": "string",
                    "title": "Type",
                    "const": categorizable.category,
                    "default": categorizable.category,
                    "readOnly": True,
                },
            }
            subclass = cls.categorizable_base_class.for_category(
                categorizable.category
            )
            if issubclass(subclass, Configurable):
                properties["config"] = {
                    "type": "object",
                    "title": "Configuration",
                    "required": subclass.required_properties,
                    "properties": subclass.properties_schema,
                }
            options.append(
                {
                    "title": categorizable.pretty_category,
                    "type": "object",
                    "properties": properties,
                }
            )
        return {
            "title": cls.pretty_name,
            "oneOf": options,
        }


# noinspection DuplicatedCode
class CategorizableBasedOptionalParameter(
    Parameter[StateType, Optional[Dict[str, Any]]],
    ABC,
    Generic[StateType, CategorizableType],
):
    async def _get(self, state: StateType) -> Optional[Dict[str, Any]]:
        categorizable = await self._get_categorizable(state)
        if categorizable is None:
            return None
        category = categorizable.category
        if isinstance(categorizable, Configurable):
            return {
                "type": category,
                "config": await categorizable.config.json.get(),
            }
        return {"type": category}

    async def _set(
        self, state: StateType, value: Optional[Dict[str, Any]]
    ) -> Callable[[], Awaitable]:
        current = await self._get_categorizable(state)

        if value is None:
            undo = await self._create_undo(state, current, None)
            await self._set_categorizable(state, None)
            if isinstance(current, Configurable):
                await current.cleanup()
            return undo

        if current is not None:
            new_category = value.get("type", current.category)
            if new_category == current.category:
                undo = await self._create_undo(state, current, current)
                await current.config.set(value.get("config", {}))
                return undo
        else:
            new_category = value["type"]

        params = await self._get_params(state, new_category)
        subclass = self.categorizable_base_class.for_category(new_category)
        if issubclass(subclass, Configurable):
            instance = await subclass.create(**params)
            await instance.config.set(value.get("config", {}))
        else:
            instance = subclass(**params)

        undo = await self._create_undo(state, current, instance)
        await self._set_categorizable(state, instance)
        if isinstance(current, Configurable):
            await current.cleanup()
        return undo

    async def _create_undo(
        self,
        state: StateType,
        old: Optional[Categorizable],
        new: Optional[Categorizable],
    ) -> Callable[[], Awaitable]:
        if old is None and new is None:
            return noop

        if old is new:
            if not isinstance(old, Configurable):
                return noop

            config = await old.config.json.get()

            async def undo():
                # noinspection PyUnresolvedReferences
                await old.config.set(config)

            return undo

        if old is None:

            async def undo():
                await self._set_categorizable(state, None)
                if isinstance(new, Configurable):
                    await new.cleanup()

            return undo

        if isinstance(old, Configurable):
            tempdir = SelfDeletingDirectory()
            await old.save(tempdir.path)
            # noinspection PyUnresolvedReferences
            params = await self._get_params(state, old.category)

            async def undo():
                loaded = await old.from_saved(tempdir.path, **params)
                await self._set_categorizable(state, loaded)
                if isinstance(new, Configurable):
                    await new.cleanup()

            return undo

        async def undo():
            await self._set_categorizable(state, old)
            if isinstance(new, Configurable):
                await new.cleanup()

        return undo

    async def _get_categorizable(
        self, state: StateType
    ) -> Optional[CategorizableType]:
        return getattr(state, decamelize(self.name))

    async def _set_categorizable(
        self, state: StateType, value: Optional[CategorizableType]
    ) -> None:
        setattr(state, decamelize(self.name), value)

    async def _get_params(
        self, state: StateType, category: str
    ) -> Dict[str, Any]:
        all_params = getattr(state, f"{decamelize(self.name)}s_params")
        return all_params.get(category, {})

    @classproperty
    def categorizable_base_class(cls) -> Type[CategorizableType]:
        return get_generic_args(cls, CategorizableBasedOptionalParameter)[1]

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        options = []
        for categorizable in cls.categorizable_base_class.all_categorizables:
            properties = {
                "type": {
                    "type": "string",
                    "title": "Type",
                    "const": categorizable.category,
                    "default": categorizable.category,
                    "readOnly": True,
                },
            }
            subclass = cls.categorizable_base_class.for_category(
                categorizable.category
            )
            if issubclass(subclass, Configurable):
                properties["config"] = {
                    "type": "object",
                    "title": "Configuration",
                    "required": subclass.required_properties,
                    "properties": subclass.properties_schema,
                }
            options.append(
                {
                    "title": categorizable.pretty_category,
                    "type": "object",
                    "properties": properties,
                }
            )
        return {
            "title": cls.pretty_name,
            "oneOf": options + [{"type": "null", "title": "None"}],
        }
